from typing import Union, List, Optional

from . import _containers
from ._ldaps import LdapObject
from ._object_classes import Container


def search(
    connection: LdapObject, search_base: Optional[str] = None, search_scope: Optional[int] = None, **search_params
) -> List[Container]:
    """ Search for an Active Directory container.

    :param connection: An LDAPS connection.
    :param search_base: An optional distinguishedName to use as the root of the search. Defaults
        to the root of the domain.
    :param search_scope: An optional int to use as the search scope. Defaults to 2 (SCOPE_SUBTREE).
    :param search_params: Keyword arguments to include in the LDAP query.

    :raises exceptions.InvalidFilter: in the unlikely case that search parameters, once automatically escaped,
        result in an invalid LDAP query.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: A list of Active Directory container objects.

    To find the container with cn 'test.cn', call the function like so:

        .. code-block:: python

            containers.search(
                connection,
                cn="test.ou"
            )

    To find all containers with a name starting with 'sales', call the function like so:

        .. code-block:: python

            containers.search(connection, name="sales*")

    .. note:: If the wildcard character is not at the end of the search string, it will be escaped.

    """
    return _containers._get(connection, search_base=search_base, search_scope=search_scope, **search_params)


def get_all(
    connection: LdapObject,
    search_base: Optional[str] = None,
    search_scope: Optional[int] = None,
    ignore_system: Optional[bool] = False,
) -> List[Container]:
    """ Get all Active Directory containers.

    :param connection: An LDAPS connection.
    :param search_base: An optional distinguishedName to use as the root of the search. Defaults
        to the root of the domain.
    :param search_scope: An optional int to use as the search scope. Defaults to 2 (SCOPE_SUBTREE).
    :param ignore_system: An optional bool to control whether the default `System` container & its
      contents should be ignored.

    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: A list of Active Directory container objects.

    Example:

        .. code-block:: python

            containers.get_all(connection)

    .. warning:: This operation can be expensive! Using the search function is recommended.

    .. seealso:: :ref:`active_directory:Search Functions`
    """
    results = _containers._get(connection, search_base=search_base, search_scope=search_scope)
    return (
        results
        if ignore_system is False
        else [c for c in results if not c.distinguishedName.endswith(f"CN=System,{connection._base_dn}")]
    )


def create(connection: LdapObject, *, distinguished_name: str, name: str, **extra_params) -> None:
    """ Create a new Active Directory container.

    :param connection: An LDAPS connection.

    :param distinguished_name: The distinguished name of the target container.
    :param name: The name for the new container.
    :param extra_params: Keyword arguments to include in the new container LDAP attributes.

    :raises exceptions.ObjectExists: if the supplied container already exists.
    :raises exceptions.ParameterError:  if an invalid LDAP attribute is supplied.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            containers.create(
                connection,
                distinguished_name='cn=test ou,cn=Users,dc=test,dc=domain,dc=com',
                name='ou name',
                info='my new ou'
            )

    .. seealso:: :ref:`active_directory:Action Functions`, :ref:`active_directory:Keyword Argument Attributes`
    """
    return _containers._create(connection, distinguished_name, name, **extra_params)


def delete(connection: LdapObject, *, distinguished_name: str) -> None:
    """ Delete a container.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target container.

    :raises exceptions.NoSuchObject: if the container does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            containers.delete(
                connection,
                distinguished_name='cn=my old ou,cn=users,dc=test,dc=domain,dc=com',
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    return _containers._delete(connection, distinguished_name)


def move(connection: LdapObject, *, distinguished_name: str, new_parent: str) -> None:
    """ Move a container.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target container.
    :param new_parent: The distinguished name of the new OU to put the container in.

    :raises exceptions.NoSuchObject: if the container or new parent does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            containers.move(
                connection,
                distinguished_name='cn=engineers,cn=users,dc=test,dc=domain,dc=com',
                new_parent='ou=staff ous,dc=test,dc=domain,dc=com'
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    return _containers._move(connection, distinguished_name, new_parent)


def set_attribute(connection: LdapObject, *, distinguished_name: str, name: str, value: Union[int, str, bytes]) -> None:
    """ Set a container LDAP attribute value.

    This function performs the LDAP `modify_add` operation. It is suitable for setting
    values against blank attributes.

    If the supplied attribute already has a value, exceptions.ModifyAttributeError
    will be raised. In this case modify_attribute should be used instead.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target container.
    :param attribute_name: The name of the LDAP attribute to modify.
    :param value: The value to set on the LDAP attribute.

    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.
    :raises exceptions.ModifyAttributeError: if there was an issue setting the new attribute value.
    :raises exceptions.NoSuchObject: if the container does not exist.

    Example:

        .. code-block:: python

            containers.set_attribute(
                connection,
                distinguished_name="cn=test ou,cn=Users,dc=test,dc=domain,dc=com",
                name="info",
                value="my new value"
            )

    .. seealso:: :ref:`active_directory:Action Functions`

    """
    _containers._modify(connection, distinguished_name, name, value, "add")


def replace_attribute(
    connection: LdapObject, *, distinguished_name: str, name: str, value: Union[int, str, bytes]
) -> None:
    """ Modify a container LDAP attribute value.

    This function performs the LDAP `modify_replace` operation.

    As suggested by the function name; if the supplied attribute already has a value, it will
    be replaced.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target container.
    :param attribute_name: The name of the LDAP attribute to modify.
    :param value: The value to replace the existing LDAP attribute value with.

    :raises exceptions.ModifyAttributeError: if there was an issue replacing the attribute value.
    :raises exceptions.NoSuchObject: if the container does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            replace_attribute(
                connection,
                distinguished_name="cn=test ou,cn=Users,dc=test,dc=domain,dc=com",
                name="info",
                value="my new value"
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    _containers._modify(connection, distinguished_name, name, value, "replace")


def delete_attribute(
    connection: LdapObject, *, distinguished_name: str, name: str, value: Union[int, str, bytes]
) -> None:
    """ Remove a container LDAP attribute value.

    This function performs the LDAP `modify_delete` operation.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target container.
    :param attribute_name: The name of the LDAP attribute to modify.
    :param value: The attribute value to delete.

    :raises exceptions.ModifyAttributeError: if there was an issue replacing the attribute value.
    :raises exceptions.NoSuchObject: if the container does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            containers.delete_attribute(
                connection,
                distinguished_name="cn=test ou,cn=Users,dc=test,dc=domain,dc=com",
                name="info",
                value="known existing value"
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    _containers._modify(connection, distinguished_name, name, value, "delete")
