import logging
import datetime
import functools

from collections import namedtuple
from typing import List, Optional

from datacaster.classes import CastDataClass

from ._translators import TranslatedTimestamp, seconds_to_units
from ._attribute_codecs import DecodedLdapString
from ._attribute_casting import generic_field_functions, generic_type_functions

logger = logging.getLogger(__name__)

Table = namedtuple("Table", ["header", "columns"])

_user_table = Table(
    header=["Common Name", "SAM Account Name", "Email Address", "Enabled", "Locked"],
    columns=["cn", "sAMAccountName", "mail", "is_enabled", "is_locked"],
)
_group_table = Table(header=["Common Name", "Description", "Type"], columns=["cn", "description", "groupType"])
_computer_table = Table(header=["Common Name", "Description"], columns=["cn", "description"])
_ou_table = Table(header=["Name", "Distinguished Name"], columns=["name", "distinguishedName"])
_cn_table = Table(
    header=["Name", "Description", "Distinguished Name"], columns=["name", "description", "distinguishedName"]
)


def order_by_samaccountname(self, other):
    if isinstance(other, self.__class__):
        return getattr(self, "sAMAccountName") < getattr(other, "sAMAccountName")
    raise ValueError(f"Cannot compare objects of {type(self)} with {type(other)}.")


# ==========================
#  Active Directory Records
#  ========================
#
# The following classes all have raw LDAP dictionaries passed to their constructors when creating new records.
#
# Every attribute value in the dictionary will be wrapped in a list, even if the attribute only has a single value.
#
# All values are bytes.
#
# When a record is created using one of these classes, each attribute with an annotation will be unpacked and cast
# to the corresponding annotated type using ONE of the following:
#
#     - Any existing __cast_attribute_name__ instance method
#     - A function in generic_field_functions that matches the attribute name
#     - A function in generic_type_functions that matches the attribute annotation
#
# The casting functions are defined in _attribute_casting.py.
#
# Any attribute present in the raw dictionary that has no annotation will be thrown away.
#
#
# =================================
# Example LDAP Attribute Dictionary
# =================================
#
# {
#     'objectClass': [b'top', b'person', b'organizationalPerson', b'user'],
#     'cn': [b'david bennett'],
#     'sn': [b'Bennett'],
#     'givenName': [b'David'],
#     'distinguishedName': [b'CN=david bennett,OU=Bulk Users,DC=2012,DC=osirium,DC=tom'],
#     'instanceType': [b'4'],
#     'whenCreated': [b'20181217083937.0Z'],
#     'whenChanged': [b'20200305161812.0Z'],
#     'uSNCreated': [b'886967'],
#     'memberOf': [
#         b'CN=Nested Level 3,CN=Users,DC=2012,DC=osirium,DC=tom',
#         b'CN=Opus Service Desk,OU=Opus Roles,DC=2012,DC=osirium,DC=tom'
#     ],
#     'uSNChanged': [b'2264255'],
#     'name': [b'david bennett'],
#     'objectGUID': [b'g\xbdo\x86\xce\x050H\xac\xf7\xablA\x8b\xcc\x0b'],
#     'userAccountControl': [b'512'],
#     'badPwdCount': [b'0'],
#     'codePage': [b'0'],
#     'countryCode': [b'0'],
#     'badPasswordTime': [b'132278784266882594'],
#     'lastLogoff': [b'0'],
#     'lastLogon': [b'132046448155766902'],
#     'pwdLastSet': [b'132278986928790498'],
#     'primaryGroupID': [b'513'],
#     'objectSid': [b'\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00]\xd1%b\xdc{b\x1c\xd0W\xe7\x90\x0b\x11\x01\x00'],
#     'adminCount': [b'1'],
#     'accountExpires': [b'9223372036854775807'],
#     'logonCount': [b'3'],
#     'sAMAccountName': [b'david.bennett'],
#     'sAMAccountType': [b'805306368'],
#     'userPrincipalName': [b'david.bennett@2012.osirium.tom'],
#     'lockoutTime': [b'0'],
#     'objectCategory': [b'CN=Person,CN=Schema,CN=Configuration,DC=2012,DC=osirium,DC=tom'],
#     'dSCorePropagationData': [b'20200220145518.0Z', b'20200204102840.0Z', b'20190206144602.0Z', b'16010101000000.0Z'],
#     'lastLogonTimestamp': [b'132161160105427431'],
#     'mail': [b'david.bennett@2012.osirium.tom']
# }


class User(CastDataClass):
    _ppa_record_name = "User"
    _ppa_table_type = _user_table

    __class_config__ = {"cast_functions": {"fields": generic_field_functions, "types": generic_type_functions}}

    accountExpires: Optional[TranslatedTimestamp]
    adminCount: Optional[int]
    badPasswordTime: Optional[TranslatedTimestamp]
    badPwdCount: Optional[int]
    carLicense: List[str]
    cn: Optional[str]
    co: Optional[str]
    comment: Optional[str]
    company: Optional[str]
    countryCode: Optional[str]
    department: Optional[str]
    description: Optional[str]
    displayName: Optional[str]
    distinguishedName: str
    division: Optional[str]
    employeeNumber: Optional[str]
    employeeType: Optional[str]
    extensionAttribute1: Optional[str]
    extensionAttribute2: Optional[str]
    extensionAttribute3: Optional[str]
    extensionAttribute4: Optional[str]
    extensionAttribute5: Optional[str]
    extensionAttribute6: Optional[str]
    extensionAttribute7: Optional[str]
    extensionAttribute8: Optional[str]
    extensionAttribute9: Optional[str]
    extensionAttribute10: Optional[str]
    extensionAttribute11: Optional[str]
    extensionAttribute12: Optional[str]
    extensionAttribute13: Optional[str]
    extensionAttribute14: Optional[str]
    extensionAttribute15: Optional[str]
    facsimileTelephoneNumber: Optional[str]
    gidNumber: Optional[int]
    givenName: Optional[str]
    info: Optional[str]
    l: Optional[str]
    lastLogoff: Optional[str]
    lastLogon: Optional[TranslatedTimestamp]
    lastLogonTimestamp: Optional[TranslatedTimestamp]
    lockoutTime: Optional[TranslatedTimestamp]
    loginShell: Optional[str]
    logonCount: Optional[int]
    logonHours: Optional[DecodedLdapString]
    mail: Optional[str]
    manager: Optional[str]
    memberOf: List[str]
    mobile: Optional[str]
    msExchRecipientDisplayType: Optional[int]
    msExchRecipientTypeDetails: Optional[int]
    msExchRemoteRecipientType: Optional[int]
    name: Optional[str]
    objectCategory: str
    objectClass: List[str]
    objectGUID: DecodedLdapString
    objectSid: DecodedLdapString
    physicalDeliveryOfficeName: Optional[str]
    postalCode: Optional[str]
    postOfficeBox: Optional[List[str]]
    primaryGroupID: Optional[int]
    pwdLastSet: Optional[TranslatedTimestamp]
    sAMAccountName: str
    sAMAccountType: Optional[int]
    sn: Optional[str]
    st: Optional[str]
    street: Optional[str]
    streetAddress: Optional[str]
    telephoneNumber: Optional[str]
    thumbnailPhoto: Optional[DecodedLdapString]
    title: Optional[str]
    uid: Optional[str]
    uidNumber: Optional[int]
    unixHomeDirectory: Optional[str]
    uSNChanged: Optional[str]
    uSNCreated: Optional[str]
    userAccountControl: int
    userPrincipalName: Optional[str]
    whenChanged: Optional[str]
    whenCreated: Optional[str]

    @property
    def is_locked(self):
        return self.lockoutTime is not None

    @property
    def is_disabled(self):
        return self.userAccountControl in {514, 522, 546}

    @property
    def is_enabled(self):
        return not self.is_disabled

    @property
    def is_expired(self):
        return datetime.datetime.utcnow() > self.accountExpires.date_object if self.accountExpires else False

    @property
    def password_age(self):
        return (
            seconds_to_units(int((datetime.datetime.utcnow() - self.pwdLastSet.date_object).total_seconds()))
            if self.pwdLastSet is not None
            else None
        )

    def __lt__(self, other):
        return order_by_samaccountname(self, other)


class Group(CastDataClass):
    _ppa_record_name = "Group"
    _ppa_table_type = _group_table

    __class_config__ = {
        "rename_fields": {"member;range=0-1499": "member_range"},
        "cast_functions": {"fields": generic_field_functions, "types": generic_type_functions},
    }

    cn: Optional[str]
    description: Optional[str]
    distinguishedName: str
    gidNumber: Optional[int]
    groupType: str
    info: Optional[str]
    managedBy: Optional[str]
    member: Optional[List[str]]  # This is populated if there are < 1500 members
    member_range: Optional[
        List[str]
    ]  # This (renamed from member;range=0-1499) is populated if there are > 1500 members
    memberUid: List[str]
    name: Optional[str]
    objectCategory: str
    objectClass: List[str]
    objectGUID: DecodedLdapString
    objectSid: DecodedLdapString
    sAMAccountName: str
    sAMAccountType: int
    uSNChanged: Optional[str]
    uSNCreated: Optional[str]
    whenChanged: Optional[str]
    whenCreated: Optional[str]

    def __lt__(self, other):
        return order_by_samaccountname(self, other)

    def __cast_groupType__(self, value):
        # https://ldapwiki.com/wiki/GroupType
        try:
            return {
                -2_147_483_646: "Global Security Group",
                -2_147_483_644: "Local Security Group",
                -2_147_483_643: "BuiltIn Group",
                -2_147_483_640: "Universal Security Group",
                2: "Global Distribution Group",
                4: "Local Distribution Group",
                8: "Universal Distribution Group",
            }[int(value[0].decode("utf-8"))]
        except KeyError:
            return f"Unknown Group Type ({value[0]})"


class Computer(CastDataClass):
    _ppa_record_name = "Computer"
    _ppa_table_type = _computer_table

    __class_config__ = {"cast_functions": {"fields": generic_field_functions, "types": generic_type_functions}}

    accountExpires: Optional[TranslatedTimestamp]
    canonicalName: Optional[str]
    carLicense: List[str]
    cn: Optional[str]
    co: Optional[str]
    comment: Optional[str]
    description: Optional[str]
    displayName: Optional[str]
    distinguishedName: str
    dNSHostName: Optional[str]
    info: Optional[str]
    lastLogoff: Optional[str]
    lastLogon: Optional[TranslatedTimestamp]
    lastLogonTimestamp: Optional[TranslatedTimestamp]
    lockoutTime: Optional[TranslatedTimestamp]
    logonCount: Optional[int]
    logonHours: Optional[DecodedLdapString]
    logonWorkstation: Optional[str]
    mail: Optional[str]
    managedBy: Optional[str]
    memberOf: List[str]
    name: Optional[str]
    o: Optional[str]
    objectCategory: str
    objectClass: List[str]
    objectGUID: DecodedLdapString
    objectSid: DecodedLdapString
    objectVersion: Optional[int]
    operatingSystem: Optional[str]
    operatingSystemHotfix: Optional[str]
    operatingSystemServicePack: Optional[str]
    operatingSystemVersion: Optional[str]
    ou: Optional[str]
    pwdLastSet: Optional[TranslatedTimestamp]
    sAMAccountName: str
    sAMAccountType: int
    uid: Optional[str]
    userAccountControl: int
    userPrincipalName: Optional[str]

    def __lt__(self, other):
        return order_by_samaccountname(self, other)


class OrganizationalUnit(CastDataClass):
    """
    {
        'objectClass': [b'top', b'organizationalUnit'],
        'ou': [b'Opus Roles'],
        'distinguishedName': [b'OU=Opus Roles,DC=2012,DC=osirium,DC=tom'],
        'instanceType': [b'4'],
        'whenCreated': [b'20190912124954.0Z'],
        'whenChanged': [b'20190930104306.0Z'],
        'uSNCreated': [b'2040930'],
        'uSNChanged': [b'2048381'],
        'name': [b'Opus Roles'],
        'objectGUID': [b'USm\x18\x92\xdf\xfaO\xa9\xa2\x0e\xc6\x90zVS'],
        'objectCategory': [b'CN=Organizational-Unit,CN=Schema,CN=Configuration,DC=2012,DC=osirium,DC=tom'],
        'dSCorePropagationData': [
            b'20200204102844.0Z',
            b'20200204102843.0Z',
            b'20190930104306.0Z',
            b'20190912124954.0Z',
            b'16010101000416.0Z'
        ]
    }

    https://docs.microsoft.com/en-gb/windows/win32/adschema/c-organizationalunit
    """

    _ppa_record_name = "OrganizationalUnit"
    _ppa_table_type = _ou_table

    __class_config__ = {"cast_functions": {"fields": generic_field_functions, "types": generic_type_functions}}

    ou: Optional[str]
    distinguishedName: str
    name: Optional[str]
    objectCategory: str
    objectClass: List[str]
    objectGUID: DecodedLdapString
    whenChanged: Optional[str]
    whenCreated: Optional[str]


class Container(CastDataClass):
    """
    {
        'objectClass': [b'top', b'container'],
        'cn': [b'Users'],
        'description': [b'Default container for upgraded user accounts'],
        'distinguishedName': [b'CN=Users,DC=2012,DC=osirium,DC=tom'],
        'instanceType': [b'4'],
        'whenCreated': [b'20180129093741.0Z'],
        'whenChanged': [b'20180129093741.0Z'],
        'uSNCreated': [b'5821'],
        'uSNChanged': [b'5821'],
        'showInAdvancedViewOnly': [b'FALSE'],
        'name': [b'Users'],
        'objectGUID': [b'\xa5v\x94\xaf\x7f=vE\xa5\xceK\xceR0\x9c\x01'],
        'systemFlags': [b'-1946157056'],
        'objectCategory': [b'CN=Container,CN=Schema,CN=Configuration,DC=2012,DC=osirium,DC=tom'],
        'isCriticalSystemObject': [b'TRUE'],
        'dSCorePropagationData': [
            b'20200204102844.0Z',
            b'20200204102843.0Z',
            b'20180223085859.0Z',
            b'20180129093942.0Z',
            b'16010714042432.0Z'
        ]
    }

    https://docs.microsoft.com/en-gb/windows/win32/adschema/c-organizationalunit
    """

    _ppa_record_name = "Container"
    _ppa_table_type = _cn_table

    __class_config__ = {"cast_functions": {"fields": generic_field_functions, "types": generic_type_functions}}

    cn: Optional[str]
    description: Optional[str]
    distinguishedName: str
    name: Optional[str]
    objectCategory: str
    objectClass: List[str]
    objectGUID: DecodedLdapString
    whenChanged: Optional[str]
    whenCreated: Optional[str]


def create_records(record_type):
    def _wrapper(func):
        @functools.wraps(func)
        def _execute(connection, **kwargs):
            return list(map(lambda ldap_attributes: record_type(**ldap_attributes), func(connection, **kwargs)))

        return _execute

    return _wrapper
