from typing import Union, List, Optional

from . import _organizational_units
from ._ldaps import LdapObject
from ._object_classes import OrganizationalUnit


def search(
    connection: LdapObject, search_base: Optional[str] = None, search_scope: Optional[int] = None, **search_params
) -> List[OrganizationalUnit]:
    """ Search for an Active Directory organizational unit.

    :param connection: An LDAPS connection.
    :param search_base: An optional distinguishedName to use as the root of the search. Defaults
        to the root of the domain.
    :param search_scope: An optional int to use as the search scope. Defaults to 2 (SCOPE_SUBTREE).
    :param search_params: Keyword arguments to include in the LDAP query.

    :raises exceptions.InvalidFilter: in the unlikely case that search parameters, once automatically escaped,
        result in an invalid LDAP query.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: A list of Active Directory organizational unit objects.

    To find the organizational unit with cn 'test.ou', call the function like so:

        .. code-block:: python

            organizational_units.search(
                connection,
                cn="test.ou"
            )

    To find all organizational units with a name starting with 'sales', call the function like so:

        .. code-block:: python

            organizational_units.search(connection, name="sales*")

    .. note:: If the wildcard character is not at the end of the search string, it will be escaped.

    """
    return _organizational_units._get(connection, search_base=search_base, search_scope=search_scope, **search_params)


def get_all(
    connection: LdapObject, search_base: Optional[str] = None, search_scope: Optional[int] = None
) -> List[OrganizationalUnit]:
    """ Get all Active Directory organizational units.

    :param connection: An LDAPS connection.
    :param search_base: An optional distinguishedName to use as the root of the search. Defaults
        to the root of the domain.
    :param search_scope: An optional int to use as the search scope. Defaults to 2 (SCOPE_SUBTREE).

    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: A list of Active Directory organizational unit objects.

    Example:

        .. code-block:: python

            organizational_units.get_all(connection)

    .. warning:: This operation can be expensive! Using the search function is recommended.

    .. seealso:: :ref:`active_directory:Search Functions`
    """
    return _organizational_units._get(connection, search_base=search_base, search_scope=search_scope)


def create(connection: LdapObject, *, distinguished_name: str, name: str, **extra_params) -> None:
    """ Create a new Active Directory organizational unit.

    :param connection: An LDAPS connection.

    :param distinguished_name: The distinguished name of the target organizational unit.
    :param name: The name for the new organizational unit.
    :param extra_params: Keyword arguments to include in the new organizational unit LDAP attributes.

    :raises exceptions.ObjectExists: if the supplied organizational unit already exists.
    :raises exceptions.ParameterError:  if an invalid LDAP attribute is supplied.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            organizational_units.create(
                connection,
                distinguished_name='cn=test ou,cn=Users,dc=test,dc=domain,dc=com',
                name='ou name',
                info='my new ou'
            )

    .. seealso:: :ref:`active_directory:Action Functions`, :ref:`active_directory:Keyword Argument Attributes`
    """
    return _organizational_units._create(connection, distinguished_name, name, **extra_params)


def delete(connection: LdapObject, *, distinguished_name: str) -> None:
    """ Delete an organizational unit.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target organizational unit.

    :raises exceptions.NoSuchObject: if the organizational unit does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            organizational_units.delete(
                connection,
                distinguished_name='cn=my old ou,cn=users,dc=test,dc=domain,dc=com',
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    return _organizational_units._delete(connection, distinguished_name)


def move(connection: LdapObject, *, distinguished_name: str, new_parent: str) -> None:
    """ Move an organizational unit.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target organizational unit.
    :param new_parent: The distinguished name of the new OU or CN to put the organizational unit in.

    :raises exceptions.NoSuchObject: if the organizational unit or new parent does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            organizational_units.move(
                connection,
                distinguished_name='cn=engineers,cn=users,dc=test,dc=domain,dc=com',
                new_parent='ou=staff ous,dc=test,dc=domain,dc=com'
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    return _organizational_units._move(connection, distinguished_name, new_parent)


def set_attribute(connection: LdapObject, *, distinguished_name: str, name: str, value: Union[int, str, bytes]) -> None:
    """ Set an organizational unit LDAP attribute value.

    This function performs the LDAP `modify_add` operation. It is suitable for setting
    values against blank attributes.

    If the supplied attribute already has a value, exceptions.ModifyAttributeError
    will be raised. In this case modify_attribute should be used instead.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target organizational unit.
    :param attribute_name: The name of the LDAP attribute to modify.
    :param value: The value to set on the LDAP attribute.

    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.
    :raises exceptions.ModifyAttributeError: if there was an issue setting the new attribute value.
    :raises exceptions.NoSuchObject: if the organizational unit does not exist.

    Example:

        .. code-block:: python

            organizational_units.set_attribute(
                connection,
                distinguished_name="cn=test ou,cn=Users,dc=test,dc=domain,dc=com",
                name="info",
                value="my new value"
            )

    .. seealso:: :ref:`active_directory:Action Functions`

    """
    _organizational_units._modify(connection, distinguished_name, name, value, "add")


def replace_attribute(
    connection: LdapObject, *, distinguished_name: str, name: str, value: Union[int, str, bytes]
) -> None:
    """ Modify an organizational unit LDAP attribute value.

    This function performs the LDAP `modify_replace` operation.

    As suggested by the function name; if the supplied attribute already has a value, it will
    be replaced.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target organizational unit.
    :param attribute_name: The name of the LDAP attribute to modify.
    :param value: The value to replace the existing LDAP attribute value with.

    :raises exceptions.ModifyAttributeError: if there was an issue replacing the attribute value.
    :raises exceptions.NoSuchObject: if the organizational unit does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            replace_attribute(
                connection,
                distinguished_name="cn=test ou,cn=Users,dc=test,dc=domain,dc=com",
                name="info",
                value="my new value"
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    _organizational_units._modify(connection, distinguished_name, name, value, "replace")


def delete_attribute(
    connection: LdapObject, *, distinguished_name: str, name: str, value: Union[int, str, bytes]
) -> None:
    """ Remove an organizational unit LDAP attribute value.

    This function performs the LDAP `modify_delete` operation.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target organizational unit.
    :param attribute_name: The name of the LDAP attribute to modify.
    :param value: The attribute value to delete.

    :raises exceptions.ModifyAttributeError: if there was an issue replacing the attribute value.
    :raises exceptions.NoSuchObject: if the organizational unit does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            organizational_units.delete_attribute(
                connection,
                distinguished_name="cn=test ou,cn=Users,dc=test,dc=domain,dc=com",
                name="info",
                value="known existing value"
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    _organizational_units._modify(connection, distinguished_name, name, value, "delete")
